'
' Operation on two arrays of 4component vectors calculated on GPU and CPU
' [!] GeForce 8 and newer, or Radeon HD 4xxx and newer with OpenCL enabled
' [!] drivers REQUIRED
'
' Code already optimized by separating the JIT compiling part
'
' Petr Schreiber, 2010
' USE AT OWN RISK
'
' Code based on fantastic C code by Yuan Wan:
' https://www.wiki.ed.ac.uk/display/ecdfwiki/OpenCL+quick+start
'

#INCLUDE "cl.inc"
#INCLUDE "logWindow.inc"

' -- Time measurement
GLOBAL t1, t2 AS DWORD, GPUTime, CPUTime AS SINGLE
DECLARE FUNCTION GetTickCount LIB "KERNEL32.DLL" ALIAS "GetTickCount" () AS DWORD

%TRUE = -1
%FALSE=  0

' -- Do you want to do error checking?
%OPENCL_ERROR_CHECKING = %FALSE

' -- Operation equates
%OPERATION_SUM      = 1
%OPERATION_SUBTRACT = 2
%OPERATION_MULTIPLY = 3
%OPERATION_DIVIDE   = 4
MACRO OPERATION_DECODE(operation) = CHOOSE$(operation, "+", "-", "*", "/")
MACRO OPERATION_DECODE_FUNC(operation) = CHOOSE$(operation, "VectorSum", "VectorSub", "VectorMul", "VectorDiv")

' -- How many element will the array have
%ELEMENTS = 2000000

' -- UDT
TYPE float4
  x AS SINGLE
  y AS SINGLE
  z AS SINGLE
  w AS SINGLE
END TYPE

FUNCTION PBMAIN()

  REGISTER i AS LONG

  LogWindow_Open()
  LogWindow_AddText("Allocating arrays ("+FORMAT$((3*%ELEMENTS*16)/1024/1024, "#,")+" MB)...", 0)
  DIM InputVectorA(%ELEMENTS-1) AS float4
  DIM InputVectorB(%ELEMENTS-1) AS float4
  DIM OutputVector(%ELEMENTS-1) AS float4
  LogWindow_AddText("Done", 1)

  LogWindow_AddText("Entering initial values...", 0)
  FOR i = 0 TO %ELEMENTS-1
    InputVectorA(i).x = 1'INT(1+RND*10)'1'rnd*80000
    InputVectorA(i).y = 2'INT(1+RND*10)'1'rnd*80000
    InputVectorA(i).z = 3'INT(1+RND*10)'1'rnd*80000
    InputVectorB(i).x = 4'INT(1+RND*10)'2'RND*80000
    InputVectorB(i).y = 5'INT(1+RND*10)'2'RND*80000
    InputVectorB(i).z = 6'INT(1+RND*10)'2'RND*80000
  NEXT
  LogWindow_AddText("Done", 1)


  LOCAL op AS LONG

  LogWindow_AddText("Allocating OpenCL...", 0)
  OpenCL_Allocate()
  LogWindow_AddText("Done", 1)

  LogWindow_AddText("Benchmarking...", -1)
  LogWindow_AddText("", 1)
  FOR op = %OPERATION_SUM TO %OPERATION_DIVIDE
    LogWindow_AddText("TEST "+OPERATION_DECODE_FUNC(op), -1)
    t1 = GetTickCount
      OpenCL_CombineArraysfloat4(op, OutputVector(), InputVectorA(), InputVectorB())
    t2 = GetTickCount
    GPUTime = (t2-t1)/1000

    LogWindow_AddText("GPU time elapsed: "+FORMAT$(GPUTime, "#.000"), -1)

    t1 = GetTickCount
      PB_CombineArraysfloat4(op, OutputVector(), InputVectorA(), InputVectorB())
    t2 = GetTickCount
    CPUTime = (t2-t1)/1000
    LogWindow_AddText("CPU time elapsed: "+FORMAT$(CPUTime, "#.000")+" "+IIF$(CPUTime > GPUTime, "-> GPU "+FORMAT$(CPUTime / GPUTime, "#.00")+"x FASTER", "-> GPU "+FORMAT$(GPUTime/CpuTime, "#.00")+"x SLOWER"), -1)
    LogWindow_AddText("", 1)
  NEXT
  LogWindow_AddText("Done", -1)

  LogWindow_AddText("Releasing OpenCL...", 0)
  OpenCL_Release()
  LogWindow_AddText("Done", 1)

  LogWindow_AddText("Press any key to quit", -1)
  LogWindow_Save(EXE.PATH$+"OpenCL_SingleVectorArrayOperations_Benchmark.bmp")
  LogWindow_AnyKey()
  LogWindow_Close()

  FUNCTION = 0
END FUNCTION

TYPE OpenCL_Info
  GPUContext AS DWORD
  GPUDevices AS DWORD PTR

  Program(4) AS DWORD
  Kernel (4) AS DWORD
  Queue  (4) AS DWORD
END TYPE

GLOBAL OpenCL AS OpenCL_Info

FUNCTION OpenCL_Allocate() AS LONG

  LOCAL errorCL AS LONG

  LOCAL platforms AS DWORD ' %CL_platform_id
  LOCAL platformsFound AS DWORD

  errorCL = clGetPlatformIDs(1, platforms, platformsFound)
  #IF %OPENCL_ERROR_CHECKING
    IF (errorCL <> %CL_SUCCESS) THEN
      MSGBOX "Failed to enumerate platforms!" +  openCL_DecodeError(errorCL)
    END IF
  #ENDIF

  LOCAL numDevices AS DWORD
  LOCAL device_ID AS DWORD
  errorCL = clGetDeviceIDs(platforms, %CL_DEVICE_TYPE_GPU, 1, device_id, numDevices)
  #IF %OPENCL_ERROR_CHECKING
    IF (errorCL <> %CL_SUCCESS) THEN
      MSGBOX "Failed to create a device group!" +  openCL_DecodeError(errorCL)
    END IF
  #ENDIF

  ' Create a context to run OpenCL on our CUDA-enabled NVIDIA GPU
  OpenCL.GPUContext = clCreateContext(0, 1, device_id, BYVAL 0, BYVAL 0, errorCL)
  #IF %OPENCL_ERROR_CHECKING
    IF (errorCL <> %CL_SUCCESS) THEN
      MSGBOX "Failed to create a context!" +  openCL_DecodeError(errorCL)
    END IF
  #ENDIF

  ' Get the list of GPU devices associated with this context
  LOCAL ParmDataBytes AS DWORD
  clGetContextInfo(OpenCL.GPUContext, %CL_CONTEXT_DEVICES, 0, BYVAL 0, ParmDataBytes)
  #IF %OPENCL_ERROR_CHECKING
    IF ParmDataBytes = 0 THEN MSGBOX "clGetContextInfo: Incorrect size of devices" : EXIT FUNCTION
  #ENDIF


  OpenCL.GPUDevices = Memory_Alloc(ParmDataBytes)
'  clGetContextInfo(GPUContext, %CL_CONTEXT_DEVICES, ParmDataBytes, GPUDevices, BYVAL 0)
  clGetContextInfo(OpenCL.GPUContext, %CL_CONTEXT_DEVICES, ParmDataBytes, BYVAL OpenCL.GPUDevices, BYVAL 0)
  #IF %OPENCL_ERROR_CHECKING
    IF GPUDevices = 0 THEN MSGBOX "clGetContextInfo: Zero GPUDevices" : EXIT FUNCTION
  #ENDIF


  DIM OpenCLSource(%OPERATION_SUM TO %OPERATION_DIVIDE) AS STRING

  LOCAL op AS LONG
  DIM kernelName(%OPERATION_SUM TO %OPERATION_DIVIDE) AS STRING


  FOR op = %OPERATION_SUM TO %OPERATION_DIVIDE
    ' Create a command-queue on the first GPU device
    OpenCL.Queue(op) = clCreateCommandQueue(OpenCL.GPUContext, OpenCL.@GPUDevices, 0, errorCL)
    #IF %OPENCL_ERROR_CHECKING
      IF errorCL THEN MSGBOX "clCreateCommandQueue: "+openCL_DecodeError(errorCL) : EXIT FUNCTION
    #ENDIF
  NEXT

  FOR op = %OPERATION_SUM TO %OPERATION_DIVIDE

    OpenCLSource(op) = "__kernel void "+OPERATION_DECODE_FUNC(op)+"(__global float4* c, __global float4* a,__global float4* b)"+$CRLF+ _
                       "{                                                                              "+$CRLF+ _
                       "  // Index of the elements to add                                              "+$CRLF+ _
                       "  unsigned int n = get_global_id(0);                                           "+$CRLF+ _
                       "                                                                               "+$CRLF+ _
                       "  // Sum the n'th element of vectors a and b and store in c                    "+$CRLF+ _
                       "  c[n] = a[n] "+OPERATION_DECODE(op)+" b[n];                                   "+$CRLF+ _
                       "}                                                                              "

    ' Create OpenCL program with source code
    LOCAL OpenCLProgram AS DWORD 'type_cl_program
    OpenCL.Program(op) = clCreateProgramWithSource(OpenCL.GPUContext, 1, OpenCLSource(op), 0, errorCL)  ' vptr
    #IF %OPENCL_ERROR_CHECKING
      IF errorCL THEN MSGBOX "clCreateProgramWithSource: "+openCL_DecodeError(errorCL) : EXIT FUNCTION
    #ENDIF

    ' Build the program (OpenCL JIT compilation)
    errorCL = clBuildProgram(OpenCL.Program(op), 0, BYVAL 0, BYVAL 0,  0, BYVAL 0)
    #IF %OPENCL_ERROR_CHECKING
      IF errorCL THEN MSGBOX "clBuildProgram: "+openCL_DecodeError(errorCL) : EXIT FUNCTION
    #ENDIF

    ' Create a handle to the compiled OpenCL function (Kernel)
    kernelName(op) = OPERATION_DECODE_FUNC(op)
    OpenCL.Kernel(op) = clCreateKernel(OpenCL.Program(op), BYCOPY kernelName(op), errorCL)
    #IF %OPENCL_ERROR_CHECKING
      IF errorCL THEN MSGBOX "clCreateKernel: "+openCL_DecodeError(errorCL) : EXIT FUNCTION
    #ENDIF

  NEXT

  FUNCTION = 0

END FUNCTION

FUNCTION OpenCL_Release() AS LONG
  LOCAL i AS LONG
  FOR i = %OPERATION_SUM TO %OPERATION_DIVIDE
    clReleaseKernel(OpenCL.Kernel(i))
    clReleaseProgram(OpenCL.Program(i))
    clReleaseCommandQueue(OpenCL.Queue(i))
  NEXT
  Memory_Free(OpenCL.GPUDevices)
  clReleaseContext(OpenCL.GPUContext)

END FUNCTION

FUNCTION OpenCL_CombineArraysfloat4(operation AS LONG, destinationArrayC() AS float4, arrayA() AS float4, arrayB() AS float4) AS LONG
  #IF %OPENCL_ERROR_CHECKING
    IF operation < %OPERATION_SUM OR operation > %OPERATION_DIVIDE THEN
      MSGBOX "Invalid operation"
      EXIT FUNCTION
    END IF

    IF UBOUND(arrayA) <> UBOUND(arrayB) OR UBOUND(arrayA) <> UBOUND(destinationArrayC) THEN
      MSGBOX "The number of dimensions of arrays must match"
      EXIT FUNCTION
    END IF

    IF LBOUND(arrayA) <> 0 OR LBOUND(arrayB) <> 0 OR LBOUND(destinationArrayC) <> 0 THEN
      MSGBOX "The lBound must be zero"
      EXIT FUNCTION
    END IF
  #ENDIF

  LOCAL errorCL AS LONG
  LOCAL uBoundOfArrays AS LONG
  LOCAL lSizeOfArrays AS LONG
  LOCAL cellSize AS LONG

  uBoundOfArrays = UBOUND(destinationArrayC)
  lSizeOfArrays  = uBoundOfArrays + 1
  cellSize       = SIZEOF(destinationArrayC(0))



  ' Allocate GPU memory for source vectors AND initialize from CPU memory
  LOCAL GPUVector1 AS DWORD 'type_cl_mem
  LOCAL GPUVector2 AS DWORD 'type_cl_mem
  GPUVector1 = clCreateBuffer(OpenCL.GPUContext, %CL_MEM_READ_ONLY OR %CL_MEM_COPY_HOST_PTR, cellSize * lSizeOfArrays, arrayA(0), errorCL)
  #IF %OPENCL_ERROR_CHECKING
    IF errorCL THEN MSGBOX "GPUVector1: "+openCL_DecodeError(errorCL) : EXIT FUNCTION
  #ENDIF

  GPUVector2 = clCreateBuffer(OpenCL.GPUContext, %CL_MEM_READ_ONLY OR %CL_MEM_COPY_HOST_PTR, cellSize * lSizeOfArrays, arrayB(0), errorCL)
  #IF %OPENCL_ERROR_CHECKING
    IF errorCL THEN MSGBOX "GPUVector2: "+openCL_DecodeError(errorCL) : EXIT FUNCTION
  #ENDIF

  ' Allocate output memory on GPU
  LOCAL GPUOutputVector AS DWORD 'type_cl_mem
  GPUOutputVector = clCreateBuffer(OpenCL.GPUContext, %CL_MEM_WRITE_ONLY, cellSize * lSizeOfArrays, BYVAL 0, errorCL)
  #IF %OPENCL_ERROR_CHECKING
    IF errorCL THEN MSGBOX "GPUOutputVector: "+openCL_DecodeError(errorCL) : EXIT FUNCTION
  #ENDIF

  ' In the next step we associate the GPU memory with the Kernel arguments
  ' This is basically setting the parameters for kernel function
  clSetKernelArg(OpenCL.Kernel(operation), 0, cellSize, GPUOutputVector)
  clSetKernelArg(OpenCL.Kernel(operation), 1, cellSize, GPUVector1)
  clSetKernelArg(OpenCL.Kernel(operation), 2, cellSize, GPUVector2)

  ' Launch the Kernel on the GPU
  DIM WorkSize AS DWORD ' one dimensional Range
  WorkSize = lSizeOfArrays
  clEnqueueNDRangeKernel(OpenCL.Queue(operation), OpenCL.Kernel(operation), 1, BYVAL 0, WorkSize, BYVAL 0, 0, BYVAL 0, BYVAL 0)

  ' Copy the output in GPU memory back to CPU memory
  clEnqueueReadBuffer(OpenCL.Queue(operation), GPUOutputVector, %CL_TRUE, 0, cellSize * lSizeOfArrays, destinationArrayC(0), 0, BYVAL 0, BYVAL 0)

  ' Cleanup
  clReleaseMemObject(GPUVector1)
  clReleaseMemObject(GPUVector2)
  clReleaseMemObject(GPUOutputVector)
  FUNCTION = 0

END FUNCTION

FUNCTION OpenCL_DecodeError(eCode AS LONG) AS STRING
  SELECT CASE eCode
    CASE 0
      FUNCTION = "No problem..."
    CASE -30
      FUNCTION = "CL_INVALID_VALUE"
    CASE -31
      FUNCTION = "CL_INVALID_DEVICE_TYPE"
    CASE -32
      FUNCTION = "CL_INVALID_PLATFORM"
    CASE -33
      FUNCTION = "CL_INVALID_DEVICE"
    CASE -34
      FUNCTION = "CL_INVALID_CONTEXT"
    CASE -35
      FUNCTION = "CL_INVALID_QUEUE_PROPERTIES"
    CASE -36
      FUNCTION = "CL_INVALID_COMMAND_QUEUE"
    CASE -37
      FUNCTION = "CL_INVALID_HOST_PTR"
    CASE -38
      FUNCTION = "CL_INVALID_MEM_OBJECT"
    CASE -39
      FUNCTION = "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR"
    CASE -40
      FUNCTION = "CL_INVALID_IMAGE_SIZE"
    CASE -41
      FUNCTION = "CL_INVALID_SAMPLER"
    CASE -42
      FUNCTION = "CL_INVALID_BINARY"
    CASE -43
      FUNCTION = "CL_INVALID_BUILD_OPTIONS"
    CASE -44
      FUNCTION = "CL_INVALID_PROGRAM"
    CASE -45
      FUNCTION = "CL_INVALID_PROGRAM_EXECUTABLE"
    CASE -46
      FUNCTION = "CL_INVALID_KERNEL_NAME"
    CASE -47
      FUNCTION = "CL_INVALID_KERNEL_DEFINITION"
    CASE -48
      FUNCTION = "CL_INVALID_KERNEL"
    CASE -49
      FUNCTION = "CL_INVALID_ARG_INDEX"
    CASE -50
      FUNCTION = "CL_INVALID_ARG_VALUE"
    CASE -51
      FUNCTION = "CL_INVALID_ARG_SIZE"
    CASE -52
      FUNCTION = "CL_INVALID_KERNEL_ARGS"
    CASE -53
      FUNCTION = "CL_INVALID_WORK_DIMENSION"
    CASE -54
      FUNCTION = "CL_INVALID_WORK_GROUP_SIZE"
    CASE -55
      FUNCTION = "CL_INVALID_WORK_ITEM_SIZE"
    CASE -56
      FUNCTION = "CL_INVALID_GLOBAL_OFFSET"
    CASE -57
      FUNCTION = "CL_INVALID_EVENT_WAIT_LIST"
    CASE -58
      FUNCTION = "CL_INVALID_EVENT"
    CASE -59
      FUNCTION = "CL_INVALID_OPERATION"
    CASE -60
      FUNCTION = "CL_INVALID_GL_OBJECT"
    CASE -61
      FUNCTION = "CL_INVALID_BUFFER_SIZE"
    CASE -62
      FUNCTION = "CL_INVALID_MIP_LEVEL"
    CASE -63
      FUNCTION = "CL_INVALID_GLOBAL_WORK_SIZE"
    CASE ELSE
      FUNCTION = "[!] Unknown error:"+STR$(eCode)
  END SELECT

END FUNCTION

FUNCTION FORMAT_CombineArraysRESULT(operation AS LONG, OutputVector() AS SINGLE, InputVectorA() AS SINGLE, InputVectorB() AS SINGLE) AS STRING
  LOCAL c AS LONG
  LOCAL sOutput AS STRING

'  FOR c = 0 TO %ELEMENTS-1
'    sOutput += FORMAT$(InputVectorA(c))+$TAB
'  NEXT
'  sOutput = RTRIM$(sOutput, $TAB)+$CRLF+OPERATION_DECODE(operation)+$CRLF
'  FOR c = 0 TO %ELEMENTS-1
'    sOutput += FORMAT$(InputVectorB(c))+$TAB
'  NEXT
'  sOutput = RTRIM$(sOutput, $TAB)+$CRLF+"="+$CRLF
'  FOR c = 0 TO %ELEMENTS-1
'    sOutput += FORMAT$(OutputVector(c))+$TAB
'  NEXT
'  sOutput = RTRIM$(sOutput, $TAB)

  FUNCTION = sOutput
END FUNCTION

FUNCTION Memory_Alloc(numberOfBytes AS LONG) AS DWORD

  LOCAL dwordPointer AS DWORD
  GLOBALMEM ALLOC numberOfBytes TO dwordPointer

  FUNCTION = dwordPointer

END FUNCTION

FUNCTION Memory_Free(dwordPointer AS DWORD) AS DWORD
  GLOBALMEM FREE dwordPointer TO dwordPointer

  FUNCTION = dwordPointer
END FUNCTION

' -- CPU

FUNCTION PB_CombineArraysfloat4(operation AS LONG, OutputVector() AS float4, InputVectorA() AS float4, InputVectorB() AS float4) AS LONG
  REGISTER i AS LONG

  #IF %OPENCL_ERROR_CHECKING
    IF operation < %OPERATION_SUM OR operation > %OPERATION_DIVIDE THEN
      MSGBOX "Invalid operation"
      EXIT FUNCTION
    END IF

    IF UBOUND(InputVectorA) <> UBOUND(InputVectorB) OR UBOUND(InputVectorA) <> UBOUND(OutputVector) THEN
      MSGBOX "The number of dimensions of arrays must match"
      EXIT FUNCTION
    END IF

    IF LBOUND(InputVectorA) <> 0 OR LBOUND(InputVectorB) <> 0 OR LBOUND(OutputVector) <> 0 THEN
      MSGBOX "The lBound must be zero"
      EXIT FUNCTION
    END IF
  #ENDIF

  LOCAL errorCL AS LONG
  LOCAL uBoundOfArrays AS LONG
  LOCAL lSizeOfArrays AS LONG

  uBoundOfArrays = UBOUND(OutputVector)
  lSizeOfArrays  = uBoundOfArrays + 1

  SELECT CASE AS LONG operation
    CASE %OPERATION_SUM
      FOR i = 0 TO uBoundOfArrays
        OutputVector(i).x = InputVectorA(i).x + InputVectorB(i).x
        OutputVector(i).y = InputVectorA(i).y + InputVectorB(i).y
        OutputVector(i).z = InputVectorA(i).z + InputVectorB(i).z
        OutputVector(i).w = InputVectorA(i).w + InputVectorB(i).w
      NEXT

    CASE %OPERATION_SUBTRACT
      FOR i = 0 TO uBoundOfArrays
        OutputVector(i).x = InputVectorA(i).x - InputVectorB(i).x
        OutputVector(i).y = InputVectorA(i).y - InputVectorB(i).y
        OutputVector(i).z = InputVectorA(i).z - InputVectorB(i).z
        OutputVector(i).w = InputVectorA(i).w - InputVectorB(i).w
      NEXT

    CASE %OPERATION_MULTIPLY
      FOR i = 0 TO uBoundOfArrays
        OutputVector(i).x = InputVectorA(i).x * InputVectorB(i).x
        OutputVector(i).y = InputVectorA(i).y * InputVectorB(i).y
        OutputVector(i).z = InputVectorA(i).z * InputVectorB(i).z
        OutputVector(i).w = InputVectorA(i).w * InputVectorB(i).w
      NEXT

    CASE %OPERATION_DIVIDE
      FOR i = 0 TO uBoundOfArrays
        OutputVector(i).x = InputVectorA(i).x / InputVectorB(i).x
        OutputVector(i).y = InputVectorA(i).y / InputVectorB(i).y
        OutputVector(i).z = InputVectorA(i).z / InputVectorB(i).z
        OutputVector(i).w = InputVectorA(i).w / InputVectorB(i).w
      NEXT
    CASE ELSE
      MSGBOX "Invalid operation:"+STR$(operation)

  END SELECT
END FUNCTION
